// W32PARAM.IDC : Automatic parameters recognition.
// version BETA2, coded by +Spath (spath@iname.com) in 01/01.
//
// This script tries to add to IDA 4.04 and older the nice "System Type"
// feature introduced in IDA 4.10. Parameters are named after a prototype
// file named API32.H, which must be placed in the same directory as IDA.EXE.
// With this script is provided a sample API32.H to which you can add 
// your own function definitions from standard C .h files. Note that a 
// wrong formated function can hang IDA, so make sure you use the same
// format as in the sample .h file, and that you have read how prototype
// parsing is done in GetAPIName/AddAPIParams. In any cas, this script 
// is quite basic, so you would always better save your IDB before 
// running it. Use, Modify, Improve, Distribute freely.

#include <idc.idc>
#define APIFILE "api32.h"

// OpenProtoFile(): open prototypes file
// returns: file descriptor of api32.h
static OpenProtoFile() {
   auto fd;

   if ((fd=fopen(APIFILE,"r"))==0) {
      Warning("W32PARAM Error: " + APIFILE + " not found.");
      return(-1);
   }
   else {
      Message("\nW32PARAM: API file " + APIFILE + " opened.");      
      return(fd);
   }
}

// GetAPIName(Prototype) : get API function name from api32.h 
// prototype string (string must be null terminated!).
// returns: API name
static GetAPIName(Prototype) {
   auto strstart,strtmp,strtot,strsub,strend;

   if ((strend=strstr(Prototype,"("))==-1) {
      return "NOTFOUND";
   }
   else {
      strtmp = strtot = 0;
      while (strtot<strend) {
         strstart = strtot;
         strsub = substr(Prototype,strtot,-1);
         strtmp = strstr(strsub,0x20);
         strtot = strtot + strtmp + 1;
      }
      return(substr(Prototype,strstart,strend));
   }
}

// AddAPIParams(Prototype,address) : add all parameters from API definition
// return: nothing
static AddAPIParams(Prototype,address) {
   auto strorg,strstart,strtmp,strend,strsub,strpar,nbparam;

   if ((strorg=strstr(Prototype,"("))==-1) {
      Message("\nW32PARAM: "+Prototype+" is wrongly defined.");
   }
   else if ((strstr(Prototype,"void"))==-1) {
      strsub = substr(Prototype,strorg,-1);
      strstart = strstr(strsub,0x20);
      nbparam = 1;

      while (strstart!=-1) {
         strend = strstr(strsub,")");
         strtmp = strstr(strsub,",");
         if (strtmp!=-1) {
            strend = strtmp;
         }
         strpar = substr(strsub,strstart+1,strend);            
         AddParamComm(address,nbparam,strpar);
         nbparam = nbparam + 1;
         strsub = substr(strsub,strend+1,-1);
         strstart = strstr(strsub,0x20);
      }
   }
}

// SearchAPIDef(APIName,fd): search API definition in prototype file
// returns: definition of the selected API name.
static SearchAPIDef(APIName,fd) {
   auto apidef,strg;

   fseek(fd,0,0);
   while ((apidef=readstr(fd))!=-1) {
      strg = GetAPIName(apidef);
      if (strg==APIName) {
         return(apidef);
      }
   }
   if (apidef==-1) {
      return(apidef);
   }
}

// GetPrevSPIns(Address): find previous instruction which pushes a dword
// Note that GetSPDiff is bugged in 4.04 and older for immediate push, 
// so we have to also check for "push" string.
// returns: instruction address
static GetPrevSPIns(Address) {
   auto ea;

   ea = Address;
   while ((GetSpDiff(ea)>-1)&&(GetMnem(PrevHead(ea,0))!="push")&&(ea!=-1)) {
      ea = PrevHead(ea,0);
   }
   if (ea!=-1) return PrevHead(ea,0);
   else return(ea);
}

// AddParamComm(Address,NbParam,Comment): direct/indirect shell for AddParamComm2
// return: nothing
static AddParamComm(Address,NbParam,Comment) {
   auto ea;

   if (IndirectCall(Address)==0) {
      AddParamComm2(Address,NbParam,Comment);
   }
   else {
      ea = FirstRef(Address);
      AddParamComm2(ea,NbParam,Comment); 
      while ((ea = NextRef(Address,ea)) != BADADDR) {
         AddParamComm2(ea,NbParam,Comment);
      }
   }
}

// AddParamComm2(Address,NbParam,Comment) : add parameter comments
// return: nothing
static AddParamComm2(Address,NbParam,Comment) {
   auto Param;
   
   Param = NbParam;
   while ((Param!=0)&&(Address!=-1)) {
      Address = GetPrevSPIns(Address);
      Param = Param - 1;
   }
   if (Address!=-1) MakeComm(Address,Comment);
}

// IndirectCall(address) : check for indirect API call
// return: -1 for indirect, 0 for direct 
static IndirectCall(address) {
   auto ea;

   if (PrevFunction(address+1)>(address-6)) return (-1);
   else return (0);
}

// FirstRef(address) : find first backward reference
// return: first code xref or first data xref
static FirstRef(address) {
   auto ea;
   if ((ea=RfirstB(address))==-1) ea=DfirstB(address);
   return(ea);
}

// NextRef(address,current) : find next backward reference
// return: next code xref or next data xref
static NextRef(address,current) {
   auto ea;
   if ((ea=RnextB(address,current))==-1) ea=DnextB(address,current);
   return(ea);
}


// FindExtSeg(): find segment with extern definitions
// return: first address of segment with extern definitions
static FindExtSeg() {
   auto ea,attrseg;

   ea = FirstSeg();
   attrseg = GetSegmentAttr(ea,SEGATTR_TYPE);
   while ((ea != BADADDR) && (attrseg != SEG_XTRN)) {
      ea = NextSeg(ea);
      attrseg = GetSegmentAttr(ea,SEGATTR_TYPE);
   }
   Message("\nW32PARAM: Extern segment is " + SegName(ea));
   return ea;
}

////////////////////////////////////////////////////////////////////////
//////////////////////////////// MAIN //////////////////////////////////
////////////////////////////////////////////////////////////////////////
static main(){
   auto ea, strg;
   auto extsegaddr, extsegend, attrseg;
   auto fd, apidef;

   // init
   extsegaddr = FindExtSeg();
   extsegend = SegEnd(extsegaddr);
   fd = OpenProtoFile();
           
   Message("\nW32PARAM: These APIs definitions were not found in "+APIFILE+" :\n");

   // main loop
   while (extsegaddr < extsegend - 4) {
      if ((strg=Name(extsegaddr))!="") {
         apidef = SearchAPIDef(strg,fd);
         if (apidef!=-1) {
            ea = FirstRef(extsegaddr);
            AddAPIParams(apidef,ea);
            while ((ea = NextRef(extsegaddr,ea)) != BADADDR) {
               AddAPIParams(apidef,ea);
            }
         }
         else {
            Message("\n "+strg);
         }
      }
      extsegaddr = extsegaddr + 4;
   }
   Message("\nW32PARAM: done.");
}
